from quart import Quart, request, render_template, redirect
from math import ceil
import peertube
import sys

# Import everything from utils.py, to keep main.py focused
# on CLI arguments parsing, and web routes declarations
from utils import *

cache = peertube.Cache()
api = peertube.API(cache)
# Let's follow subscriptions from the current working dir
subscriptions = Subscriptions('.', api, cache)

# --- WEBSERVER STUFF ---
# All utilities are above this point.
# Routes are defined below
app = Quart(__name__)

# --- INDEX ROUTE ---
@app.route("/")
async def main():
    info, info_failures = subscriptions.info()
    videos, videos_failures = subscriptions.videos(limit=12)
    # Inside subscriptions variable, you may find either an account info structure, or a channel info structure. Channels may be recognized due to `ownerAccount` property.

    return await render_template(
        "index.html",
        videos=videos,
        subscriptions=info,
        videos_failures=format_errors_html(videos_failures),
        info_failures=format_errors_html(info_failures),
    )
# --- END INDEX ROUTE ---

# --- SEARCH ROUTES ---
@app.route("/search", methods = ["POST"])
async def search_post_redirect():
    query = (await request.form)["query"]
    return redirect("/search/" + query)

@app.route("/search", methods = ["GET"])
async def search_get_redirect():
    query = request.args.get("query")
    return redirect("/search/" + query)

@app.route("/<string:domain>/search", methods=["POST"])
async def search_instance_redirect(domain):
    query = (await request.form)["query"]
    return redirect("/" + domain + "/search/" + query)

@app.route("/search/<string:query>", defaults = {"page": 1})
@app.route("/search/<string:query>/<int:page>")
async def search(query, page):
    results = api.search(query, (page - 1) * 10)
    return await render(
        "simpleer_search_results.html",
        results,
        lambda res: {
            "results": res,
            "query": query,
            "page": page,
            "pages_total": ceil(len(res) / 10),
        }
    )

@app.route("/<string:domain>/search/<string:term>", defaults = {"page": 1})
@app.route("/<string:domain>/search/<string:term>/<int:page>")
async def search_instance(domain, term, page):
    results = api.search_instance(domain, term, (page - 1) * 10)
    instance_name = api.instance_name(domain)
    return await render(
        "search_results.html",
        (results, instance_name),
        lambda res: {
            "domain": domain,
            "results": res[0],
            "instance": res[1],
            "search_term": term,
            # details for pagination
            "page": page,
            "pagination_url": "/" + domain + "/search/" + term + "/",
            "pages_total": ceil(len(res) / 10),

        }
    )
# --- END SEARCH ROUTES ---

# --- INSTANCE ROUTES ---
@app.route("/<string:domain>/")
async def instance(domain):
    # favicon.ico is not a domain name
    if domain == "favicon.ico": return await favicon()
    return redirect("/" + domain + "/videos/trending")

@app.route("/<string:domain>/videos/<string:filt>", defaults = {"page": 1})
@app.route("/<string:domain>/videos/<string:filt>/<int:page>")
async def instance_videos(domain, filt, page):
    if filt not in [ "most-liked", "local", "trending", "recently-added" ]:
        return await render_custom_error("No such video category: %s" % filt, 404)

    results = api.instance_videos(domain, filt=filt, start=(page - 1) * 10)
    instance_name = api.instance_name(domain)
    
    return await render(
        "instance/%s.html" % filt,
        (results, instance_name),
        lambda res: {
            "domain": domain,
            "videos": res[0],
            "instance": res[1],
            # details for pagination
            "page": page,
            "pagination_url": "/" + domain + "/videos/" + filt + "/",
            "pages_total": ceil(len(res) / 10),
        }
    )
# --- END INSTANCE ROUTES ---

# --- VIDEO ROUTE ---
@app.route("/<string:domain>/videos/watch/<string:id>/")
async def video(domain, id):
    data = api.video(domain, id)
    data["captions"] = api.video_captions(domain, id)

    quality = request.args.get("quality")
    embed = request.args.get("embed")
    vid = VideoWrapper(data, quality)
    quality = int(vid.quality)

    # Comments can be either a string containing all formatted comment, or a struct
    # containing an `exception` field when the request failed for some reason
    comments = format_comments(api.video_comments(domain, id)) if data["commentsEnabled"] else ""

    return await render(
        "video.html",
        data,
        lambda res: {
            "domain": domain,
            "instance": api.instance_name(domain),
            "video": vid,
            "comments": comments,
            "quality": quality,
            "embed": embed,
        }
    )

@app.route("/<string:domain>/videos/watch/<string:id>/<string:lang>.vtt")
async def captions(domain, id, lang):
    captions = api.video_captions(domain, id)
    if isinstance(captions, FailedRequest):
        return await render_error(captions)

    for entry in captions:
        if entry["language"]["id"] == lang:
            content = api.video_captions_proxy(domain, entry["captionPath"].split('/')[-1])
            return await render_plaintext(content)

    return await render_custom_error("This video has no subtitles/captions in requested language", code=404)
# --- END VIDEO ROUTE ---

# --- ACCOUNT ROUTES ---
@app.route("/<string:domain>/accounts/<string:name>")
async def accounts_redirect(domain, name):
    return redirect("/" + domain + "/accounts/" + name + "/video-channels")

@app.route("/<string:domain>/accounts/<string:name>/video-channels", defaults = {"page": 1})
@app.route("/<string:domain>/accounts/<string:name>/video-channels/<int:page>")
async def account_channels(domain, name, page):
    results = api.account_channels(domain, name, (page - 1) * 10)
    instance_name = api.instance_name(domain)
    account = api.account(domain, name)
    return await render(
        "accounts/channels.html",
        (results, instance_name, account),
        lambda res: {
            "domain": domain,
            "name": name,
            "video_channels": res[0],
            "instance": res[1],
            "account": res[2],
            # details for pagination
            "page": page,
            "pagination_url": "/" + domain + "/accounts/" + name + "/video-channels/",
            "pages_total": ceil(len(res) / 10),
        }
    )

@app.route("/<string:domain>/accounts/<string:name>/videos", defaults = {"page": 1})
@app.route("/<string:domain>/accounts/<string:name>/videos/<int:page>")
async def account_videos(domain, name, page):
    results = api.account_videos(domain, name, (page - 1) * 10)
    instance = api.instance_name(domain)
    account = api.account(domain, name)
    return await render(
        "accounts/videos.html",
        (results, instance, account),
        lambda res: {
            "domain": domain,
            "name": name,
            "videos": res[0],
            "instance": res[1],
            "account": res[2],
            # details for pagination
            "page": page,
            "pagination_url": "/" + domain + "/accounts/"  + name + "/videos/",
            "pages_total": ceil(len(res) / 10),
        }
    )

@app.route("/<string:domain>/accounts/<string:name>/about")
async def account_about(domain, name):
    results = api.account(domain, name)
    instance = api.instance_name(domain)
    return await render(
        "accounts/about.html",
        (results, instance),
        lambda res: {
            "domain": domain,
            "name": name,
            "account": res[0],
            "instance": res[1],
        }
    )
# --- END ACCOUNT ROUTES ---

# --- CHANNEL ROUTES ---
@app.route("/<string:domain>/video-channels/<string:name>")
async def video_channels_redirect(domain, name):
    return redirect("/" + domain + "/video-channels/" + name + "/videos")

@app.route("/<string:domain>/video-channels/<string:name>/videos", defaults = {"page": 1})
@app.route("/<string:domain>/video-channels/<string:name>/videos/<int:page>")
async def channel_videos(domain, name, page):
    results = api.channel_videos(domain, name, (page - 1) * 10)
    channel = api.channel(domain, name)
    instance = api.instance_name(domain)
    if isinstance(channel, FailedRequest):
        account = FailedRequest("Previous channel info request failed.")
    else:
        account = api.account(channel["ownerAccount"]["host"], channel["ownerAccount"]["name"])
    return await render(
        "video_channels/videos.html",
        (results, channel, instance, account),
        lambda res: {
            "domain": domain,
            "name": name,
            "videos": res[0],
            "video_channel": res[1],
            "instance": res[2],
            "account": res[3],
            # details for pagination
            "page": page,
            "pagination_url": "/" + domain + "/video-channels/" + name + "/videos/",
            "pages_total": ceil(len(res) / 10),
        }
    )

@app.route("/<string:domain>/video-channels/<string:name>/video-playlists", defaults = {"page": 1})
@app.route("/<string:domain>/video-channels/<string:name>/video-playlists/<int:page>")
async def channel_playlists(domain, name, page):
    results = api.channel_playlists(domain, name, (page - 1) * 10)
    instance = api.instance_name(domain)
    channel = api.channel(domain, name)
    if isinstance(channel, FailedRequest):
        account = FailedRequest("Previous channel info request failed.")
    else:
        account = api.account(channel["ownerAccount"]["host"], channel["ownerAccount"]["name"])
    return await render(
        "video_channels/video_playlists.html",
        (results, instance, channel, account),
        lambda res: {
            "domain": domain,
            "name": name,
            "video_playlists": res[0],
            "instance": res[1],
            "video_channel": res[2],
            "account": res[3],
            # details for pagination
            "page": page,
            "pagination_url": "/" + domain + "/accounts/"  + name + "/videos/",
            "pages_total": ceil(len(res) / 10),
        }
    )

@app.route("/<string:domain>/video-channels/<string:name>/about")
async def channel_about(domain, name):
    results = api.channel(domain, name)
    instance_name = api.instance_name(domain)
    return await render(
        "video_channels/about.html",
        (results, instance_name),
        lambda res: {
            "domain": domain,
            "name": name,
            "video_channel": res[0],
            "instance": res[1],
        }
    )

# --- DIVERSE ROUTES ---
@app.route("/favicon.ico")
async def favicon():
    return await render_custom_error("We don't have a favicon yet. If you would like to contribute one, please send it to ~metalune/public-inbox@lists.sr.ht", code=404)

@app.route("/opensearch.xml")
async def opensearch():
    try:
        with open('opensearch.xml', 'r') as f:
            return f.read().replace('$BASEURL', request.headers["Host"])
    except Exception as e:
        return await render_error(FailedRequest(e))
# --- END DIVERSE ROUTES ---
# --- END WEBSERVER STUFF ---

# --- CLI STUFF ---
if __name__ == "__main__":
    if len(sys.argv) == 3:
        interface = sys.argv[1]
        port = sys.argv[2]
    elif len(sys.argv) == 2:
        interface = "127.0.0.1"
        port = sys.argv[1]
    else:
        interface = "127.0.0.1"
        port = "5000"
    app.run(host=interface, port=port)
# --- END CLI STUFF ---
